from dataclasses import dataclass
from typing import Iterator, Optional


@dataclass(frozen=True, repr=False)
class Batch:
    size: int
    x_read_slice: Optional[slice]
    y_read_slice: Optional[slice]
    x_buffer_slice: slice
    y_buffer_slice: slice

    def __repr__(self) -> str:
        x, y = self.x_buffer_slice, self.y_buffer_slice
        s = f"Batch({self.size}, x[{x.start}:{x.stop}], y[{y.start}:{y.stop}]"
        if self.x_read_slice:
            s += f", x_read[{self.x_read_slice.start}:{self.x_read_slice.stop}]"
        if self.y_read_slice:
            s += f", y_read[{self.y_read_slice.start}:{self.y_read_slice.stop}]"
        return s + ")"


def iter_batches(
    x_buffer_size: int,
    y_buffer_size: int,
    start_offset: int,
    stop_offset: int,
) -> Iterator[Batch]:
    """
    Generate `Batch` instances describing each batch.

    Each yielded `Batch` instance describes:
    - Its size
    - The slice to read from the x array (if the current x buffer is consumed).
    - The slice to read from the y array (if the current y buffer is consumed).
    - The batch slice to read from the x buffer.
    - The batch slice to read from the y buffer.

    :param x_buffer_size: (Max) size of the x buffer.
    :param y_buffer_size: (Max) size of the y buffer.
    :param start_offset: Start row offset.
    :param stop_offset: Stop row offset.
    """
    x_read_slices = iter_slices(start_offset, stop_offset, x_buffer_size)
    y_read_slices = iter_slices(start_offset, stop_offset, y_buffer_size)
    x_buf_offset = x_buffer_size
    y_buf_offset = y_buffer_size
    offset = start_offset
    while offset < stop_offset:
        if x_buf_offset == x_buffer_size:
            x_read_slice = next(x_read_slices)
            x_read_size = x_read_slice.stop - x_read_slice.start
            x_buf_offset = 0
        else:
            x_read_slice = None

        if y_buf_offset == y_buffer_size:
            y_read_slice = next(y_read_slices)
            y_read_size = y_read_slice.stop - y_read_slice.start
            y_buf_offset = 0
        else:
            y_read_slice = None

        buffer_size = min(x_read_size - x_buf_offset, y_read_size - y_buf_offset)
        x_next_buf_offset = x_buf_offset + buffer_size
        y_next_buf_offset = y_buf_offset + buffer_size
        yield Batch(
            buffer_size,
            x_read_slice,
            y_read_slice,
            slice(x_buf_offset, x_next_buf_offset),
            slice(y_buf_offset, y_next_buf_offset),
        )
        x_buf_offset = x_next_buf_offset
        y_buf_offset = y_next_buf_offset
        offset += buffer_size


def iter_slices(start: int, stop: int, step: int) -> Iterator[slice]:
    offsets = range(start, stop, step)
    yield from map(slice, offsets, offsets[1:])
    yield slice(offsets[-1], stop)



"""Tests for TileDB PyTorch model save and load."""

import inspect
import os
import platform
import sys

import pytest
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optimizers

from tiledb.ml.models.pytorch import PyTorchTileDBModel


class Net(nn.Module):
    def __init__(self):
        super(Net, self).__init__()
        # 1 input image channel, 6 output channels, 3x3 square convolution
        # kernel
        self.conv1 = nn.Conv2d(1, 6, 3)
        self.conv2 = nn.Conv2d(6, 16, 3)
        # an affine operation: y = Wx + b
        self.fc1 = nn.Linear(16 * 6 * 6, 120)  # 6*6 from image dimension
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        # Max pooling over a (2, 2) window
        x = F.max_pool2d(F.relu(self.conv1(x)), (2, 2))
        # If the size is a square you can only specify a single number
        x = F.max_pool2d(F.relu(self.conv2(x)), 2)
        x = x.view(-1, self.num_flat_features(x))
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x

    def num_flat_features(self, x):
        size = x.size()[1:]  # all dimensions except the batch dimension
        num_features = 1
        for s in size:
            num_features *= s
        return num_features


class SeqNeuralNetwork(nn.Module):
    def __init__(self):
        super(SeqNeuralNetwork, self).__init__()
        self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(14 * 14, 256),
            nn.ReLU(),
            nn.Linear(256, 256),
            nn.ReLU(),
            nn.Linear(256, 10),
            nn.ReLU(),
        )

    def forward(self, x):
        x = self.flatten(x)
        logits = self.linear_relu_stack(x)
        return logits


class ConvNet(nn.Module):
    def __init__(self):
        super(ConvNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 6, 5)
        self.pool = nn.MaxPool2d(2, 2)
        self.conv2 = nn.Conv2d(6, 16, 5)
        self.fc1 = nn.Linear(16 * 5 * 5, 120)
        self.fc2 = nn.Linear(120, 84)
        self.fc3 = nn.Linear(84, 10)

    def forward(self, x):
        x = self.pool(F.relu(self.conv1(x)))
        x = self.pool(F.relu(self.conv2(x)))
        x = x.view(-1, 16 * 5 * 5)
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x


@pytest.mark.parametrize(
    "optimizer",
    [
        getattr(optimizers, name)
        for name, obj in inspect.getmembers(optimizers)
        if inspect.isclass(obj) and name != "Optimizer"
    ],
)
@pytest.mark.parametrize(
    "net",
    [
        getattr(sys.modules[__name__], name)
        for name, obj in inspect.getmembers(sys.modules[__name__])
        if inspect.isclass(obj) and obj.__module__ == __name__
    ],
)
class TestPyTorchModel:
    def test_save(self, tmpdir, net, optimizer):
        EPOCH = 5
        LOSS = 0.4
        model = net()
        saved_optimizer = optimizer(model.parameters(), lr=0.001)
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(
            uri=tiledb_array, model=model, optimizer=saved_optimizer
        )

        tiledb_obj.save(
            model_info={
                "epoch": EPOCH,
                "loss": LOSS,
            }
        )
        loaded_net = net()
        loaded_optimizer = optimizer(loaded_net.parameters(), lr=0.001)
        model_out_dict = tiledb_obj.load(model=loaded_net, optimizer=loaded_optimizer)

        assert model_out_dict["epoch"] == EPOCH
        assert model_out_dict["loss"] == LOSS

        # Check model parameters
        for key_item_1, key_item_2 in zip(
            model.state_dict().items(), loaded_net.state_dict().items()
        ):
            assert torch.equal(key_item_1[1], key_item_2[1])

        # Check optimizer parameters
        for key_item_1, key_item_2 in zip(
            saved_optimizer.state_dict().items(), loaded_optimizer.state_dict().items()
        ):
            assert all([a == b for a, b in zip(key_item_1[1], key_item_2[1])])

    def test_preview(self, tmpdir, net, optimizer):
        # With model given as argument
        model = net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(uri=tiledb_array, model=model)
        assert type(tiledb_obj.preview()) == str
        tiledb_obj_none = PyTorchTileDBModel(uri=tiledb_array, model=None)
        assert tiledb_obj_none.preview() == ""

    def test_file_properties(self, tmpdir, net, optimizer):
        model = net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(uri=tiledb_array, model=model)

        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK"] == "PYTORCH"
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_STAGE"] == "STAGING"
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_PYTHON_VERSION"]
            == platform.python_version()
        )
        assert (
            tiledb_obj._file_properties["TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION"]
            == torch.__version__
        )
        assert tiledb_obj._file_properties["TILEDB_ML_MODEL_PREVIEW"] == str(model)


class TestPyTorchModelCloud:
    def test_get_cloud_uri_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = Net()
        uri = os.path.join(tmpdir, "model_array")

        mock_get_cloud_uri = mocker.patch(
            "tiledb.ml.models.base.get_cloud_uri", return_value=uri
        )

        _ = PyTorchTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_cloud_uri.assert_called_once_with(uri, "test_namespace")

    def test_get_s3_prefix_call_for_models_on_tiledb_cloud(self, tmpdir, mocker):
        model = Net()
        uri = os.path.join(tmpdir, "model_array")

        mock_get_s3_prefix = mocker.patch(
            "tiledb.ml.models._cloud_utils.get_s3_prefix", return_value="s3 prefix"
        )

        _ = PyTorchTileDBModel(uri=uri, namespace="test_namespace", model=model)

        mock_get_s3_prefix.assert_called_once_with("test_namespace")

    def test_update_file_properties_call(self, tmpdir, mocker):
        model = Net()
        uri = os.path.join(tmpdir, "model_array")

        mocker.patch("tiledb.ml.models.base.get_cloud_uri", return_value=uri)

        tiledb_obj = PyTorchTileDBModel(
            uri=uri, namespace="test_namespace", model=model
        )

        mock_update_file_properties = mocker.patch(
            "tiledb.ml.models.pytorch.update_file_properties", return_value=None
        )
        mocker.patch("tiledb.ml.models.pytorch.PyTorchTileDBModel._write_array")

        tiledb_obj.save()

        file_properties_dict = {
            "TILEDB_ML_MODEL_ML_FRAMEWORK": "PYTORCH",
            "TILEDB_ML_MODEL_ML_FRAMEWORK_VERSION": torch.__version__,
            "TILEDB_ML_MODEL_STAGE": "STAGING",
            "TILEDB_ML_MODEL_PYTHON_VERSION": platform.python_version(),
            "TILEDB_ML_MODEL_PREVIEW": str(model),
        }

        mock_update_file_properties.assert_called_once_with(uri, file_properties_dict)

    def test_exception_raise_file_property_in_meta_error(self, tmpdir):
        model = Net()
        tiledb_array = os.path.join(tmpdir, "model_array")
        tiledb_obj = PyTorchTileDBModel(uri=tiledb_array, model=model)
        with pytest.raises(ValueError) as ex:
            tiledb_obj.save(
                meta={"TILEDB_ML_MODEL_ML_FRAMEWORK": "TILEDB_ML_MODEL_ML_FRAMEWORK"},
            )

        assert "Please avoid using file property key names as metadata keys!" in str(
            ex.value
        )



"""Functionality for loading data from TileDB arrays to the Tensorflow Data API."""

from functools import partial
from typing import Iterator, Optional, Sequence, Union

import sparse
import tensorflow as tf

import tiledb

from ._buffer_utils import get_attr_names, get_buffer_size
from ._tensor_gen import TileDBSparseTensorGenerator, tensor_generator

# TODO: We have to track the following issues:
# - https://github.com/tensorflow/tensorflow/issues/47532
# - https://github.com/tensorflow/tensorflow/issues/47931
# - https://github.com/tensorflow/tensorflow/issues/44565


def TensorflowTileDBDataset(
    x_array: tiledb.Array,
    y_array: tiledb.Array,
    *,
    batch_size: int,
    buffer_bytes: Optional[int] = None,
    shuffle_buffer_size: int = 0,
    prefetch: int = tf.data.AUTOTUNE,
    x_attrs: Sequence[str] = (),
    y_attrs: Sequence[str] = (),
) -> tf.data.Dataset:
    """Return a tf.data.Dataset for loading data from TileDB arrays.

    :param x_array: TileDB array of the features.
    :param y_array: TileDB array of the labels.
    :param batch_size: Size of each batch.
    :param buffer_bytes: Maximum size (in bytes) of memory to allocate for reading from
        each array (default=`tiledb.default_ctx().config()["sm.memory_budget"]`).
    :param shuffle_buffer_size: Number of elements from which this dataset will sample.
    :param prefetch: Maximum number of batches that will be buffered when prefetching.
        By default, the buffer size is dynamically tuned.
    :param x_attrs: Attribute names of x_array.
    :param y_attrs: Attribute names of y_array.
    """
    # Check that x_array and y_array have the same number of rows
    rows: int = x_array.shape[0]
    if rows != y_array.shape[0]:
        raise ValueError("X and Y arrays must have the same number of rows")

    if not x_attrs:
        x_attrs = get_attr_names(x_array.schema)
    if not y_attrs:
        y_attrs = get_attr_names(y_array.schema)

    dataset = tf.data.Dataset.from_generator(
        partial(
            tensor_generator,
            x_array=x_array,
            y_array=y_array,
            x_buffer_size=get_buffer_size(x_array, x_attrs, buffer_bytes),
            y_buffer_size=get_buffer_size(y_array, y_attrs, buffer_bytes),
            x_attrs=x_attrs,
            y_attrs=y_attrs,
            sparse_generator_cls=TensorflowSparseTensorGenerator,
        ),
        output_signature=(
            *_iter_tensor_specs(x_array.schema, x_attrs),
            *_iter_tensor_specs(y_array.schema, y_attrs),
        ),
    )
    dataset = dataset.unbatch()
    if shuffle_buffer_size > 0:
        dataset = dataset.shuffle(shuffle_buffer_size)
    return dataset.batch(batch_size).prefetch(prefetch)


def _iter_tensor_specs(
    schema: tiledb.ArraySchema, attrs: Sequence[str]
) -> Iterator[Union[tf.TensorSpec, tf.SparseTensorSpec]]:
    cls = tf.SparseTensorSpec if schema.sparse else tf.TensorSpec
    for attr in attrs:
        yield cls(shape=(None, *schema.shape[1:]), dtype=schema.attr(attr).dtype)


class TensorflowSparseTensorGenerator(TileDBSparseTensorGenerator[tf.SparseTensor]):
    @staticmethod
    def _tensor_from_coo(coo: sparse.COO) -> tf.SparseTensor:
        return tf.SparseTensor(coo.coords.T, coo.data, coo.shape)



"""Functionality for loading data from TileDB arrays to the PyTorch Dataloader API."""

import itertools as it
import math
import random
from typing import Any, Callable, Iterable, Iterator, Optional, Sequence, TypeVar

import numpy as np
import sparse
import torch

import tiledb

from ._buffer_utils import get_attr_names, get_buffer_size
from ._tensor_gen import TileDBSparseTensorGenerator, tensor_generator


def PyTorchTileDBDataLoader(
    x_array: tiledb.Array,
    y_array: tiledb.Array,
    *,
    batch_size: int,
    buffer_bytes: Optional[int] = None,
    shuffle_buffer_size: int = 0,
    prefetch: int = 2,
    x_attrs: Sequence[str] = (),
    y_attrs: Sequence[str] = (),
    num_workers: int = 0,
) -> torch.utils.data.DataLoader:
    """Return a DataLoader for loading data from TileDB arrays.

    :param x_array: TileDB array of the features.
    :param y_array: TileDB array of the labels.
    :param batch_size: Size of each batch.
    :param buffer_bytes: Maximum size (in bytes) of memory to allocate for reading
        from each array (default=`tiledb.default_ctx().config()["sm.memory_budget"]`).
    :param prefetch: Number of samples loaded in advance by each worker. Not applicable
        (and should not be given) when `num_workers` is 0.
    :param shuffle_buffer_size: Number of elements from which this dataset will sample.
    :param x_attrs: Attribute names of x_array.
    :param y_attrs: Attribute names of y_array.
    :param num_workers: how many subprocesses to use for data loading
    """
    x_schema = x_array.schema
    y_schema = y_array.schema
    if not x_attrs:
        x_attrs = get_attr_names(x_schema)
    if not y_attrs:
        y_attrs = get_attr_names(y_schema)

    return torch.utils.data.DataLoader(
        dataset=PyTorchTileDBDataset(
            x_array, y_array, buffer_bytes, shuffle_buffer_size, x_attrs, y_attrs
        ),
        batch_size=batch_size,
        prefetch_factor=prefetch,
        num_workers=num_workers,
        collate_fn=CompositeCollator(
            (
                np_arrays_collate
                if not is_sparse
                else torch.utils.data.dataloader.default_collate
            )
            for is_sparse in it.chain(
                it.repeat(x_schema.sparse, len(x_attrs)),
                it.repeat(y_schema.sparse, len(y_attrs)),
            )
        ),
    )


class PyTorchTileDBDataset(torch.utils.data.IterableDataset[Sequence[torch.Tensor]]):
    def __init__(
        self,
        x_array: tiledb.Array,
        y_array: tiledb.Array,
        buffer_bytes: Optional[int] = None,
        shuffle_buffer_size: int = 0,
        x_attrs: Sequence[str] = (),
        y_attrs: Sequence[str] = (),
    ):
        super().__init__()
        rows: int = x_array.shape[0]
        if rows != y_array.shape[0]:
            raise ValueError("X and Y arrays must have the same number of rows")

        if not x_attrs:
            x_attrs = get_attr_names(x_array.schema)
        if not y_attrs:
            y_attrs = get_attr_names(y_array.schema)

        self._rows = rows
        self._shuffle_buffer_size = shuffle_buffer_size
        self._generator_kwargs = dict(
            x_array=x_array,
            y_array=y_array,
            x_buffer_size=get_buffer_size(x_array, x_attrs, buffer_bytes),
            y_buffer_size=get_buffer_size(y_array, y_attrs, buffer_bytes),
            x_attrs=x_attrs,
            y_attrs=y_attrs,
            sparse_generator_cls=PyTorchSparseTensorGenerator,
        )

    def __iter__(self) -> Iterator[Sequence[torch.Tensor]]:
        kwargs = self._generator_kwargs
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is not None:
            for array_key in "x_array", "y_array":
                if isinstance(kwargs[array_key], tiledb.SparseArray):
                    raise NotImplementedError(
                        "https://github.com/pytorch/pytorch/issues/20248"
                    )
            per_worker = int(math.ceil(self._rows / worker_info.num_workers))
            start_offset = worker_info.id * per_worker
            stop_offset = min(start_offset + per_worker, self._rows)
            kwargs = dict(start_offset=start_offset, stop_offset=stop_offset, **kwargs)

        rows: Iterator[Sequence[torch.Tensor]] = (
            row
            for batch_tensors in tensor_generator(**kwargs)
            for row in zip(*batch_tensors)
        )
        if self._shuffle_buffer_size > 0:
            rows = iter_shuffled(rows, self._shuffle_buffer_size)
        return rows


class CompositeCollator:
    """
    A callable for collating "rows" of data into Tensors.

    Each data "column" is collated to a torch.Tensor by a different collator function.
    Finally, the collated columns are returned as a sequence of torch.Tensors.
    """

    def __init__(self, collators: Iterable[Callable[[Sequence[Any]], torch.Tensor]]):
        self._collators = tuple(collators)

    def __call__(self, rows: Sequence[Sequence[Any]]) -> Sequence[torch.Tensor]:
        columns = list(zip(*rows))
        assert len(columns) == len(self._collators)
        return [collator(column) for collator, column in zip(self._collators, columns)]


def np_arrays_collate(arrays: Sequence[np.ndarray]) -> torch.Tensor:
    # Specialized version of default_collate for collating Numpy arrays
    # Faster than `torch.as_tensor(arrays)` (https://github.com/pytorch/pytorch/pull/51731)
    # and `torch.stack([torch.as_tensor(array) for array in arrays]])`
    return torch.as_tensor(np.stack(arrays))


T = TypeVar("T")


def iter_shuffled(iterable: Iterable[T], buffer_size: int) -> Iterator[T]:
    """
    Shuffle the given iterable with a buffer.

    The buffer with `buffer_size` is filled with elements from the iterable first.
    Then, each item will be yielded from the buffer by reservoir sampling via iterator.

    """
    iterator = iter(iterable)
    buffer = list(it.islice(iterator, buffer_size))
    randrange = random.randrange
    for x in iterator:
        idx = randrange(0, buffer_size)
        yield buffer[idx]
        buffer[idx] = x
    random.shuffle(buffer)
    while buffer:
        yield buffer.pop()


class PyTorchSparseTensorGenerator(TileDBSparseTensorGenerator[torch.Tensor]):
    @staticmethod
    def _tensor_from_coo(coo: sparse.COO) -> torch.Tensor:
        return torch.sparse_coo_tensor(coo.coords, coo.data, coo.shape)



import itertools as it
import os
import uuid
from contextlib import contextmanager

import numpy as np
import pytest
import tensorflow as tf
import torch

import tiledb


def parametrize_for_dataset(
    x_sparse=(True, False),
    y_sparse=(True, False),
    x_shape=((10,), (10, 3)),
    y_shape=((5,), (5, 2)),
    num_attrs=(1, 2),
    pass_attrs=(True, False),
    batch_size=(8,),
    buffer_bytes=(1024, None),
    shuffle_buffer_size=(0, 16),
):
    argnames = [
        "x_sparse",
        "y_sparse",
        "x_shape",
        "y_shape",
        "num_attrs",
        "pass_attrs",
        "buffer_bytes",
        "batch_size",
        "shuffle_buffer_size",
    ]
    argvalues = it.product(
        x_sparse,
        y_sparse,
        x_shape,
        y_shape,
        num_attrs,
        pass_attrs,
        buffer_bytes,
        batch_size,
        shuffle_buffer_size,
    )
    return pytest.mark.parametrize(argnames, argvalues)


@contextmanager
def ingest_in_tiledb(
    tmpdir,
    x_data,
    y_data,
    x_sparse,
    y_sparse,
    num_attrs,
    pass_attrs,
):
    """Context manager for ingest data into TileDB.

    Yield the keyword arguments for instantiating a TiledbDataset.
    """
    array_uuid = str(uuid.uuid4())
    x_uri = os.path.join(tmpdir, "x_" + array_uuid)
    y_uri = os.path.join(tmpdir, "y_" + array_uuid)
    _ingest_in_tiledb(x_uri, x_data, x_sparse, num_attrs)
    _ingest_in_tiledb(y_uri, y_data, y_sparse, num_attrs)
    attrs = [f"features_{attr}" for attr in range(num_attrs)] if pass_attrs else []
    with tiledb.open(x_uri) as x_array, tiledb.open(y_uri) as y_array:
        yield dict(x_array=x_array, y_array=y_array, x_attrs=attrs, y_attrs=attrs)


def _ingest_in_tiledb(uri: str, data: np.ndarray, sparse: bool, num_attrs: int) -> None:
    dims = [
        tiledb.Dim(
            name=f"dim_{dim}",
            domain=(0, data.shape[dim] - 1),
            tile=np.random.randint(1, data.shape[dim] + 1),
            dtype=np.int32,
        )
        for dim in range(data.ndim)
    ]

    # TileDB schema
    schema = tiledb.ArraySchema(
        domain=tiledb.Domain(*dims),
        sparse=sparse,
        attrs=[
            tiledb.Attr(name=f"features_{attr}", dtype=np.float32)
            for attr in range(num_attrs)
        ],
    )

    # Create the (empty) array on disk.
    tiledb.Array.create(uri, schema)

    # Ingest
    with tiledb.open(uri, "w") as tiledb_array:
        idx = np.nonzero(data) if sparse else slice(None)
        tiledb_array[idx] = {f"features_{attr}": data[idx] for attr in range(num_attrs)}


def rand_array(num_rows: int, *row_shape: int, sparse: bool = False) -> np.ndarray:
    """Create a random array of shape (num_rows, *row_shape).

    :param num_rows: Number of rows of the array (i.e. first dimension size).
    :param row_shape: Shape of each row (i.e. remaining dimension sizes).
    :param sparse:
      - If false, all values will be in the (0, 1) range.
      - If true, only `num_rows` values will be in the (0, 1) range, the rest will be 0.
        Note: some rows may be all zeros.
    """
    shape = (num_rows, *row_shape)
    if sparse:
        a = np.zeros(shape)
        flat_idxs = np.random.choice(a.size, size=len(a), replace=False)
        a.flat[flat_idxs] = np.random.random(len(flat_idxs))
    else:
        a = np.random.random(shape)
    assert a.shape == shape
    return a


def validate_tensor_generator(
    generator, num_attrs, x_sparse, y_sparse, x_shape, y_shape, batch_size=None
):
    for tensors in generator:
        assert len(tensors) == 2 * num_attrs
        # the first num_attrs tensors are the features (x)
        for tensor in tensors[:num_attrs]:
            _validate_tensor(tensor, x_sparse, x_shape, batch_size)
        # the last num_attrs tensors are the labels (y)
        for tensor in tensors[num_attrs:]:
            _validate_tensor(tensor, y_sparse, y_shape, batch_size)


def _validate_tensor(tensor, expected_sparse, expected_shape, batch_size=None):
    if batch_size is None:
        row_shape = tensor.shape
    else:
        num_rows, *row_shape = tensor.shape
        # num_rows may be less than batch_size
        assert num_rows <= batch_size, (num_rows, batch_size)
    assert tuple(row_shape) == expected_shape
    assert _is_sparse_tensor(tensor) == expected_sparse


def _is_sparse_tensor(tensor):
    if isinstance(tensor, np.ndarray):
        return False
    if isinstance(tensor, torch.Tensor):
        return tensor.is_sparse
    if isinstance(tensor, tf.SparseTensor):
        return True
    if isinstance(tensor, tf.Tensor):
        return False
    assert False, f"Unknown tensor type: {type(tensor)}"



from abc import ABC, abstractmethod
from typing import Generic, Iterator, Sequence, Type, TypeVar, Union

import numpy as np
import sparse

import tiledb

from ._batch_utils import iter_batches

Tensor = TypeVar("Tensor")


class TileDBNumpyGenerator:
    """Base class for generating tensors read from a TileDB array."""

    def __init__(self, array: tiledb.Array, attrs: Sequence[str]) -> None:
        """
        :param array: TileDB array to read from.
        :param attrs: Attribute names of array to read.
        """
        self._query = array.query(attrs=attrs)

    def read_buffer(self, array_slice: slice) -> None:
        """
        Read an array slice and save it as the current buffer.

        :param array_slice: Requested array slice.
        """
        self._buf_arrays = tuple(self._query[array_slice].values())

    def iter_tensors(self, buffer_slice: slice) -> Iterator[np.ndarray]:
        """
        Return an iterator of tensors for the given slice, one tensor per attribute

        Must be called after `read_buffer`.

        :param buffer_slice: Slice of the current buffer to convert to tensors.
        """
        return (buf_array[buffer_slice] for buf_array in self._buf_arrays)


class TileDBSparseTensorGenerator(TileDBNumpyGenerator, ABC, Generic[Tensor]):
    def __init__(self, array: tiledb.Array, attrs: Sequence[str]) -> None:
        self._dims = tuple(array.domain.dim(i).name for i in range(array.ndim))
        self._row_shape = array.shape[1:]
        super().__init__(array, attrs)

    def read_buffer(self, array_slice: slice) -> None:
        buffer = self._query[array_slice]
        coords = [buffer.pop(dim) for dim in self._dims]
        # normalize the first coordinate dimension to start at start_offset
        start_offset = array_slice.start
        if start_offset:
            coords[0] -= start_offset
        shape = (array_slice.stop - start_offset, *self._row_shape)
        self._buf_arrays = tuple(
            sparse.COO(coords, data, shape) for data in buffer.values()
        )

    def iter_tensors(self, buffer_slice: slice) -> Iterator[Tensor]:
        return map(self._tensor_from_coo, super().iter_tensors(buffer_slice))

    @staticmethod
    @abstractmethod
    def _tensor_from_coo(coo: sparse.COO) -> Tensor:
        """Convert a sparse.COO to a Tensor"""


def tensor_generator(
    x_array: tiledb.Array,
    y_array: tiledb.Array,
    x_buffer_size: int,
    y_buffer_size: int,
    x_attrs: Sequence[str],
    y_attrs: Sequence[str],
    sparse_generator_cls: Type[TileDBSparseTensorGenerator[Tensor]],
    start_offset: int = 0,
    stop_offset: int = 0,
) -> Iterator[Sequence[Union[np.ndarray, Tensor]]]:
    """
    Generator for batches of tensors.

    Each yielded batch is a sequence of N tensors of x_array followed by M tensors
    of y_array, where `N == len(x_attrs)` and `M == len(y_attrs)`.

    :param x_array: TileDB array of the features.
    :param y_array: TileDB array of the labels.
    :param x_buffer_size: Number of rows to read at a time from x_array.
    :param y_buffer_size: Number of rows to read at a time from y_array.
    :param x_attrs: Attribute names of x_array.
    :param y_attrs: Attribute names of y_array.
    :param start_offset: Start row offset; defaults to 0.
    :param stop_offset: Stop row offset; defaults to number of rows.
    :param sparse_generator_cls: Sparse tensor generator type.
    """
    x_gen: Union[TileDBNumpyGenerator, TileDBSparseTensorGenerator[Tensor]] = (
        sparse_generator_cls(x_array, x_attrs)
        if x_array.schema.sparse
        else TileDBNumpyGenerator(x_array, x_attrs)
    )
    y_gen: Union[TileDBNumpyGenerator, TileDBSparseTensorGenerator[Tensor]] = (
        sparse_generator_cls(y_array, y_attrs)
        if y_array.schema.sparse
        else TileDBNumpyGenerator(y_array, y_attrs)
    )
    if not stop_offset:
        stop_offset = x_array.shape[0]
    for batch in iter_batches(x_buffer_size, y_buffer_size, start_offset, stop_offset):
        if batch.x_read_slice:
            x_gen.read_buffer(batch.x_read_slice)
        if batch.y_read_slice:
            y_gen.read_buffer(batch.y_read_slice)
        x_tensors = x_gen.iter_tensors(batch.x_buffer_slice)
        y_tensors = y_gen.iter_tensors(batch.y_buffer_slice)
        yield (*x_tensors, *y_tensors)
